/*
 *	$Id: tcpsocket.c,v 1.2 2005/05/02 23:44:21 rh Exp $
 *
 * A simple TCP socket class
 *
 */

#define	_POSIX_C_SOURCE	200112	/* required for fdopen() */
#define __EXTENSIONS__		/* required on Solaris */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/param.h>
#include <sys/socket.h>

#include "tcpsocket.h"

#define BACKLOG 4		/* number of pending connections */

#ifdef BSD			/* is this a BSD system? */
#define	ON_BSD(code)	code	/* then compile BSD specific code */
#else
#define	ON_BSD(code)		/* on other systems, just ignore this */

#if !defined(socklen_t) && !defined(__socklen_t_defined)
typedef int socklen_t;		/* define our own socklen_t if not defined */
#endif

#endif /* if/else BSD */

/*
 * create a socket that connects to a server
 *
 * parameters:
 *	- host: the host to connect to
 *	- port: the port to connect to
 *
 * return values:
 *	a pointer to an allocated TCPSocket
 */
TCPSocket *new_TCPSocket(char *host, int port)
{
	TCPSocket *sock = TCPSocket_alloc();

	if (sock == NULL) return NULL;

	if (!TCPSocket_init(sock))
	{
		TCPSocket_free(sock);
		return NULL;
	}

	return TCPSocket_connect(sock, host, port) ? sock : NULL;
}

/*
 * allocate memory for a TCPSocket
 *
 * parameters:
 *	none
 *
 * return values:
 *	a pointer to an allocated TCPSocket
 */
TCPSocket *TCPSocket_alloc(void)
{
	return calloc(sizeof(TCPSocket), 1);
}


/*
 * release a created socket
 *
 * parameters:
 *	- sock: the socket to release
 *
 * return values:
 *	none
 */
void TCPSocket_free(TCPSocket *sock)
{
	if (sock->have_stream)
	{
		fclose(sock->in);
		fclose(sock->out);
	}
	else if (sock->is_open) close(sock->s);

	free(sock);
}


/*
 * initialise a socket
 *
 * parameters:
 *	- sock: the socket to initialise
 *
 * return values:
 *	1 if successful, 0 otherwise
 */
int TCPSocket_init(TCPSocket *sock)
{
	sock->s = socket(PF_INET, SOCK_STREAM, 0);

	if (sock->s < 0) return 0;

	return sock->is_open = 1;		/* socket is open */
}


/*
 * connect a socket to a server
 *
 * parameters:
 *	- sock: the socket to connect
 *	- host: the host to connect to
 *	- port: the port to connect to
 *
 * return values:
 *	1 if successful, 0 otherwise
 */
int TCPSocket_connect(TCPSocket *sock, char *host, int port)
{
	struct hostent *he = gethostbyname(host);

	if (he == NULL) return 0;		/* can't find host? */

	ON_BSD(sock->remote.sin_len = sizeof(sock->remote));
	sock->remote.sin_family = AF_INET;	/* internet socket */
	sock->remote.sin_port = htons(port);	/* the port to connect to */
	sock->remote.sin_addr = *((struct in_addr *)he->h_addr);

	if (connect(sock->s, (struct sockaddr *)&sock->remote,
					sizeof(sock->remote)) < 0)
		return 0;			/* cannot connect */

	return sock->is_connected = 1;		/* connection established */
}

/*
 * establish buffered I/O on a socket
 *
 * parameters:
 *	- sock: the socket to operate on
 *
 * return values:
 *	1 if successful, 0 otherwise;
 */
int TCPSocket_createStream(TCPSocket *sock)
{
	int sout = dup(sock->s);	/* duplicate the socket for output */

	if (sout == -1) return 0;	/* cannot duplicate socket */

	sock->in = fdopen(sock->s, "r");/* create a buffered reader */

	if (sock->in == NULL)
	{
		close(sout);
		return 0;
	}

	sock->out = fdopen(sout, "w");	/* create a buffered writer */

	if (sock->out == NULL)
	{
		fclose(sock->in);	/* this closes sock->s as well */
		sock->s = sout;		/* so use the dup'ed one from now on */
		return 0;
	}

	return sock->have_stream = 1;	/* we have a stream now */
}


/*
 * get the buffered input stream of the current socket (open if necessary)
 *
 * parameters:
 *	- sock: a connected socket to operate on
 *
 * return values:
 *	an input stream FILE * if successful, NULL otherwise
 */
FILE *TCPSocket_getInputStream(TCPSocket *sock)
{
	if (!sock->have_stream && !TCPSocket_createStream(sock)) return NULL;

	return sock->in;
}


/*
 * get the buffered output stream of the current socket (open if necessary)
 *
 * parameters:
 *	- sock: a connected socket object to operate on
 *
 * return values:
 *	an input stream FILE * if successful, NULL otherwise
 */
FILE *TCPSocket_getOutputStream(TCPSocket *sock)
{
	if (!sock->have_stream && !TCPSocket_createStream(sock)) return NULL;

	return sock->out;
}


/*
 * get the low level socket (only possible if no I/O streams have been opened)
 *
 * parameters:
 *	- sock: a socket object to operate on
 *
 * return values:
 *	a socket descriptor if successful, -1 otherwise
 */
int TCPSocket_getSocket(TCPSocket *sock)
{
	return sock->have_stream ? -1 : sock->s;
}
